#pragma once
#ifndef BOOTLOADER_SETTINGS_H
#define BOOTLOADER_SETTINGS_H

// Copied the relevant parts from NRF SDK 15.2.0 nrf_dfu_types.h

#include "stdtypes.h"

#define INIT_COMMAND_MAX_SIZE   256 /**< Maximum size of the init command stored in dfu_settings. */
#define SYSTEM_SERVICE_ATT_SIZE 8   /**< Size of the system service attribute length including CRC-16 at the end. */

#pragma pack(4)
typedef struct
{
   uint32_t                image_size;         /**< Size of the image in the bank. */
   uint32_t                image_crc;          /**< CRC of the image. If set to 0, the CRC is ignored. */
   uint32_t                bank_code;          /**< Identifier code for the bank. */
} nrf_dfu_bank_t;

typedef struct
{
   uint32_t command_size;              /**< The size of the current init command stored in the DFU settings. */
   uint32_t command_offset;            /**< The offset of the currently received init command data. The offset will increase as the init command is received. */
   uint32_t command_crc;               /**< The calculated CRC of the init command (calculated after the transfer is completed). */
   uint32_t data_object_size;          /**< The size of the last object created. Note that this size is not the size of the whole firmware image.*/
   union
   {
      struct
      {
         uint32_t firmware_image_crc;        /**< CRC value of the current firmware (continuously calculated as data is received). */
         uint32_t firmware_image_crc_last;   /**< The CRC of the last executed object. */
         uint32_t firmware_image_offset;     /**< The offset of the current firmware image being transferred. Note that this offset is the offset in the entire firmware image and not only the current object. */
         uint32_t firmware_image_offset_last;/**< The offset of the last executed object from the start of the firmware image. */
      };
      struct
      {
         uint32_t update_start_address;      /**< Value indicating the start address of the new firmware (before copy). It's always used, but it's most important for an SD/SD+BL update where the SD changes size or if the DFU process had a power loss when updating a SD with changed size. */
      };
   };
} dfu_progress_t;

typedef struct
{
   uint32_t            crc;                                    /**< CRC of the rest of the parameters in this struct. */
   ble_gap_id_key_t    ble_id;                                 /**< BLE GAP identity key of the device that initiated the DFU process. */
   ble_gap_enc_key_t   enc_key;                                /**< Encryption key structure containing encrypted diversifier and LTK for reestablishing the bond. */
   uint8_t             sys_serv_attr[SYSTEM_SERVICE_ATT_SIZE]; /**< System service attributes for restoring of Service Changed Indication setting in DFU mode. */
} nrf_dfu_peer_data_t;

typedef struct
{
   uint32_t    crc;                            /**< CRC of the rest of the parameters in this struct. Calculated by the bootloader. */
   uint8_t     name[20];                       /**< New advertisement name to set. */
   uint32_t    len;                            /**< Length of the advertisement name. */
} nrf_dfu_adv_name_t;

typedef struct
{
   uint32_t            crc;                /**< CRC for the stored DFU settings, not including the CRC itself. If 0xFFFFFFF, the CRC has never been calculated. */
   uint32_t            settings_version;   /**< Version of the current DFU settings struct layout. */
   uint32_t            app_version;        /**< Version of the last stored application. */
   uint32_t            bootloader_version; /**< Version of the last stored bootloader. */

   uint32_t            bank_layout;        /**< Bank layout: single bank or dual bank. This value can change. */
   uint32_t            bank_current;       /**< The bank that is currently used. */

   nrf_dfu_bank_t      bank_0;             /**< Bank 0. */
   nrf_dfu_bank_t      bank_1;             /**< Bank 1. */

   uint32_t            app_offset;         /**< for direct writes within application space - used for Renishaw manufacturer-specific-data flash area */
   uint32_t            write_offset;       /**< Write offset for the current operation. */
   uint32_t            sd_size;            /**< Size of the SoftDevice. */

   dfu_progress_t      progress;           /**< Current DFU progress. */

   uint32_t            enter_buttonless_dfu;
   uint8_t             init_command[INIT_COMMAND_MAX_SIZE];  /**< Buffer for storing the init command. */

   nrf_dfu_peer_data_t peer_data;          /**< Not included in calculated CRC. */
   nrf_dfu_adv_name_t  adv_name;           /**< Not included in calculated CRC. */

} nrf_dfu_settings_t;
#pragma pack() // revert pack settings

#endif // BOOTLOADER_SETTINGS_H
