#include "ble_stack.h"
#include "flash.h"

#include "app_error.h"

//IAR Specific
extern uint32_t  __ICFEDIT_region_RAM_start__;
uint32_t const * const m_ram_start = &__ICFEDIT_region_RAM_start__;

#define APP_RAM_START   (uint32_t)m_ram_start

#define IS_GAP_EVT(id)        (bool)((id >= BLE_GAP_EVT_BASE) && (id <= BLE_GAP_EVT_LAST))
#define IS_GATTS_EVT(id)      (bool)((id >= BLE_GATTS_EVT_BASE) && (id <= BLE_GATTS_EVT_LAST))
#define IS_GATTC_EVT(id)      (bool)((id >= BLE_GATTC_EVT_BASE) && (id <= BLE_GATTC_EVT_LAST))

#define NUMBER_OF_GAP_EVT     (uint8_t)(BLE_GAP_EVT_LAST - BLE_GAP_EVT_BASE)
#define NUMBER_OF_GATTS_EVT   (uint8_t)(BLE_GATTS_EVT_LAST - BLE_GATTS_EVT_BASE)
#define NUMBER_OF_GATTC_EVT   (uint8_t)(BLE_GATTC_EVT_LAST - BLE_GATTC_EVT_BASE)

#define MAX_SUBSCRIBERS       2
#define NOT_SUPPORTED         NULL

#define SCAN_INTERVAL           0x00A0                                  /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW             0x0050                                  /**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_DURATION           0x0000                                  /**< Timout when scanning. 0x0000 disables timeout. */

#define MIN_CONNECTION_INTERVAL MSEC_TO_UNITS(7.5, UNIT_1_25_MS)         /**< Determines minimum connection interval in millisecond. */
#define MAX_CONNECTION_INTERVAL MSEC_TO_UNITS(7.5, UNIT_1_25_MS)         /**< Determines maximum connection interval in millisecond. */
#define SLAVE_LATENCY           0                                       /**< Determines slave latency in counts of connection events. */
#define SUPERVISION_TIMEOUT     MSEC_TO_UNITS(4000, UNIT_10_MS)         /**< Determines supervision time-out in units of 10 millisecond. */

#define ADVERT_REPORT_BUFFER_SIZE  255

/** @brief buffer used to store advert result from scanning. */
static uint8_t adv_report_buffer[ADVERT_REPORT_BUFFER_SIZE];
ble_data_t scan_buffer= 
{
adv_report_buffer,
ADVERT_REPORT_BUFFER_SIZE
};

/** @brief Parameters used when scanning. */
static ble_gap_scan_params_t const m_scan_params =
{
    .active   = 1,
    .extended = 0,
    .interval = SCAN_INTERVAL,
    .window   = SCAN_WINDOW,
    .timeout          = SCAN_DURATION,
    .scan_phys        = BLE_GAP_PHY_1MBPS,
    .filter_policy    = BLE_GAP_SCAN_FP_ACCEPT_ALL,
};

/**@brief Connection parameters requested for connection. */
static ble_gap_conn_params_t const m_connection_param =
{
    (uint16_t)MIN_CONNECTION_INTERVAL,  // Minimum connection
    (uint16_t)MAX_CONNECTION_INTERVAL,  // Maximum connection
    (uint16_t)SLAVE_LATENCY,            // Slave latency
    (uint16_t)SUPERVISION_TIMEOUT       // Supervision time-out
};

/* Helpers */
__STATIC_INLINE uint32_t fnBleEnable(uint32_t * const p_app_ram_start);
__STATIC_INLINE void fnConfigureBleStack(const ble_stack_config_t * const config);
__STATIC_INLINE void fnInitGapParams();
__STATIC_INLINE void fnProcessEvent(const ble_evt_t * const p_ble_evt);
__STATIC_INLINE ble_stack_evt_handler_t fnGetEventHandler(uint16_t id, uint32_t level);
__STATIC_INLINE bool fnSetEventHandler(const uint16_t id, const uint32_t level, const ble_stack_evt_handler_t handler);
__STATIC_INLINE uint32_t fnBleStack_DataLengthUpdate(const uint16_t conn_handle, uint16_t data_length);

/* Default GAP Event Handlers*/
__STATIC_INLINE void fnOnConnected(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnConnectionParamsUpdated(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnDisconnected(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnConnParamUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt); // NEW
__STATIC_INLINE void fnOnSecurityParamsRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnDataLengthUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnDataLengthUpdate(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnPhyUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnPhyUpdate(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnTimeout(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);

/* Default GATT Client Event Handler */
__STATIC_INLINE void fnOnWriteResponse(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnMtuResponse(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);


static ble_stack_evt_handler_t gfpGapEventHandlers[NUMBER_OF_GAP_EVT][MAX_SUBSCRIBERS] =
{
   { fnOnConnected, NULL },                   //BLE_GAP_EVT_CONNECTED
   { fnOnDisconnected, NULL },                //BLE_GAP_EVT_DISCONNECTED
   { fnOnConnectionParamsUpdated, NULL },     //BLE_GAP_EVT_CONN_PARAM_UPDATE
   { fnOnSecurityParamsRequest, NULL },       //BLE_GAP_EVT_SEC_PARAMS_REQUEST
   { NULL, NULL },                            //BLE_GAP_EVT_SEC_INFO_REQUEST
   { NULL, NULL },                            //BLE_GAP_EVT_PASSKEY_DISPLAY
   { NULL, NULL },                            //BLE_GAP_EVT_KEY_PRESSED
   { NULL, NULL },                            //BLE_GAP_EVT_AUTH_KEY_REQUEST
   { NULL, NULL },                            //BLE_GAP_EVT_LESC_DHKEY_REQUEST
   { NULL, NULL },                            //BLE_GAP_EVT_AUTH_STATUS
   { NULL, NULL },                            //BLE_GAP_EVT_CONN_SEC_UPDATE
   { fnOnTimeout, NULL },                     //BLE_GAP_EVT_TIMEOUT
   { NULL, NULL },                            //BLE_GAP_EVT_RSSI_CHANGED
   { NULL, NULL },                            //BLE_GAP_EVT_ADV_REPORT
   { NULL, NULL },                            //BLE_GAP_EVT_SEC_REQUEST
   { fnOnConnParamUpdateRequest, NULL },      //BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST
   { NULL, NULL },                            //BLE_GAP_EVT_SCAN_REQ_REPORT
   { fnOnPhyUpdateRequest, NULL },            //BLE_GAP_EVT_PHY_UPDATE_REQUEST
   { fnOnPhyUpdate, NULL },                   //BLE_GAP_EVT_PHY_UPDATE
   { fnOnDataLengthUpdateRequest, NULL },     //BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST
   { fnOnDataLengthUpdate, NULL },            //BLE_GAP_EVT_DATA_LENGTH_UPDATE
   { NULL, NULL },                            //BLE_GAP_EVT_QOS_CHANNEL_SURVEY_REPORT
   { NULL, NULL }                             //BLE_GAP_EVT_ADV_SET_TERMINATED
};

static ble_stack_evt_handler_t gfpGattcEventHandlers[NUMBER_OF_GATTC_EVT][MAX_SUBSCRIBERS] = 
{
    { NULL, NULL },                           //BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_REL_DISC_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_CHAR_DISC_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_DESC_DISC_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_ATTR_INFO_DISC_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_READ_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_CHAR_VALS_READ_RSP
    { fnOnWriteResponse, NULL },              //BLE_GATTC_EVT_WRITE_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_HVX
    { fnMtuResponse, NULL },                  //BLE_GATTC_EVT_EXCHANGE_MTU_RSP
    { NULL, NULL },                           //BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE
};

// from sd 6.1.0 - global scope for these is important
static uint8_t gabyEncodedAdvData[BLE_GAP_ADV_SET_DATA_SIZE_MAX] = { 0 };
static uint8_t gabyEncodedScanData[BLE_GAP_ADV_SET_DATA_SIZE_MAX] = { 0 };
static ble_gap_adv_data_t  gstGapAdvData = { 0 };

static uint8_t ghAdvertising = BLE_GAP_ADV_SET_HANDLE_NOT_SET;
static uint16_t ghConnection = BLE_CONN_HANDLE_INVALID;          /**< Handle of the current connection. */

static ble_stack_params_t gstDefaultStackParams =
{
   (uint32_t)(MSEC_TO_UNITS(187.5, UNIT_0_625_MS)),   /**< The advertising interval (in units of 0.625 ms. This value corresponds to 187.5 ms). */
   (uint16_t)(MSEC_TO_UNITS(600000, UNIT_10_MS)),     /**< The advertising duration (600 seconds) in units of 10 milliseconds. */
   (uint16_t)(MSEC_TO_UNITS(7.5, UNIT_1_25_MS)),      /* CONN_INTERVAL_DEFAULT */
   (uint16_t)(MSEC_TO_UNITS(7.5, UNIT_1_25_MS)),      /* EVENT_LENGTH_DEFAULT */
   (uint8_t)1,                                        /* EXTENDED_EVENT_LENGTH  1: enable, 0: disable */
   (uint16_t)0,                                       /* SLAVE_LATENCY */
   (uint16_t)MSEC_TO_UNITS(1000, UNIT_10_MS)          /* CONN_SUP_TIMEOUT */
};

static ble_stack_params_t *gpstBleParams;

///
void fnBleStack_Initialize(const ble_stack_config_t * const config)
{
   fnConfigureBleStack(config);
   fnInitGapParams();
}

///
bool fnBleStack_Subscribe(const uint32_t eventId, const ble_stack_evt_handler_t ble_evt_handler)
{
   const uint32_t level = 1;

   APP_ASSERT(fnGetEventHandler(eventId, level) == NULL);
   return fnSetEventHandler(eventId, level, ble_evt_handler);
}

///
bool fnBleStack_Unsubscribe(const uint32_t eventId)
{
   return fnSetEventHandler(eventId, 1, NULL);
}

///
void fnBleStack_PollEvents(void)
{
   while (1)
   {
      uint32_t err_code;
      __ALIGN(4) uint8_t evt_buffer[NRF_SDH_BLE_GATT_MAX_MTU_SIZE] = { 0 }; // TODO:: Stack usage ???

      ble_evt_t * p_ble_evt;
      uint16_t    evt_len = (uint16_t)sizeof(evt_buffer);

      err_code = sd_ble_evt_get(evt_buffer, &evt_len);

      if (NRF_SUCCESS == err_code)
      {
          p_ble_evt = (ble_evt_t*)evt_buffer;
          fnProcessEvent(p_ble_evt);
      }
      else if (err_code == NRF_ERROR_NOT_FOUND)
      {
          // no more events available
          break;
      }
      else
      {
          //NRF_ERROR_INVALID_ADDR
          //NRF_ERROR_DATA_SIZE 
          LOG_DEBUG("ERROR on BLE stack event retrieval %d \r\n", err_code);
      }
   }
}
///
uint32_t fnBleStack_SendNotification(const uint16_t value_handle, const uint8_t * const p_data, uint16_t * p_length)
{
   if (!fnBleStack_IsConnected())
      return BLE_ERROR_INVALID_CONN_HANDLE;
      
   ble_gatts_hvx_params_t const hvx_param =
   {
      .type = BLE_GATT_HVX_NOTIFICATION,
      .handle = value_handle,
      .p_data = p_data,
      .p_len = p_length,
   };
   
   uint32_t err_code = sd_ble_gatts_hvx(ghConnection, &hvx_param);

   if (NRF_ERROR_RESOURCES == err_code)
   {
      LOG_DEBUG("sd_ble_gatts_hvx() NRF_ERROR_RESOURCES\r\n");
   }
   else if (NRF_SUCCESS != err_code)
   {
      LOG_DEBUG("sd_ble_gatts_hvx() failed: 0x%x\r\n", err_code);
   }

   return err_code;
}

///
bool fnBleStack_IsConnected(void)
{
   return (BLE_CONN_HANDLE_INVALID != ghConnection);
}

///
void fnBleStack_DropConnection(void)
{
   if (!fnBleStack_IsConnected())
      return;

   uint32_t err_code = sd_ble_gap_disconnect(ghConnection, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
   if (NRF_SUCCESS != err_code)
   {
      LOG_DEBUG("sd_ble_gap_disconnect() failed: 0x%x\r\n", err_code);
   }
}

///
bool fnBleStack_SetStackParameters(const ble_stack_params_t * const p_params)
{
   flash_write_operation_t *op = fnFlash_BeginWrite((MSD_FLASH_START_ADDR + FLASH_BLE_PARAMS_T_OFFSET), (uint8_t const*)p_params, FLASH_BLE_PARAMS_T_LENGTH);
   fnFlash_UpdateWriteChecksum(op);
   return fnFlash_EndWrite(op);
}

uint32_t fnBleStack_SetGattsValue(uint16_t service_handle, ble_gatts_value_t * p_value)
{
    return sd_ble_gatts_value_set(ghConnection, service_handle, p_value);
}

uint32_t fnBleStack_ScanStart()
{
    uint32_t err_code = 0;
    
    err_code = sd_ble_gap_scan_start(&m_scan_params, &scan_buffer);
  
    return err_code;
}

uint32_t fnBleStack_ScanResume()
{
    uint32_t err_code = 0;
    
    err_code = sd_ble_gap_scan_start(NULL, &scan_buffer);
  
    return err_code;
}

uint32_t fnBleStack_Connect(ble_gap_addr_t const *p_peer_addr)
{
    uint32_t err_code = 0;
    
    err_code = sd_ble_gap_connect(p_peer_addr, &m_scan_params, &m_connection_param, APP_BLE_CONN_CFG_TAG);
    
    return err_code;
}

uint32_t fnBleStack_PrimaryServicesDiscover(uint16_t handle, ble_uuid_t const *  UUID)
{
   uint32_t err_code = 0;
  
   err_code = sd_ble_gattc_primary_services_discover(ghConnection, handle, UUID);

   return err_code;
}

uint32_t fnBleStack_GattcWrite(ble_gattc_write_params_t const * write_params)
{
   uint32_t err_code = 0;
  
   err_code = sd_ble_gattc_write(ghConnection, write_params);

   return err_code;
}

__STATIC_INLINE void fnConfigureBleStack(const ble_stack_config_t * const config)
{
   ble_opt_t  opt;
   ble_cfg_t ble_cfg;

   uint32_t ram_start = APP_RAM_START;

   gpstBleParams = &gstDefaultStackParams;

   // Configure the connection count.
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.conn_cfg.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;
   ble_cfg.conn_cfg.params.gap_conn_cfg.conn_count = NRF_SDH_BLE_TOTAL_LINK_COUNT;
   ble_cfg.conn_cfg.params.gap_conn_cfg.event_length = gpstBleParams->default_event_length;

   uint32_t err_code = sd_ble_cfg_set(BLE_CONN_CFG_GAP, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the connection roles.
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.gap_cfg.role_count_cfg.periph_role_count = NRF_SDH_BLE_PERIPHERAL_LINK_COUNT;
   ble_cfg.gap_cfg.role_count_cfg.central_role_count = NRF_SDH_BLE_CENTRAL_LINK_COUNT;
   ble_cfg.gap_cfg.role_count_cfg.central_sec_count = MIN(NRF_SDH_BLE_CENTRAL_LINK_COUNT, BLE_GAP_ROLE_COUNT_CENTRAL_SEC_DEFAULT);

   err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the maximum ATT MTU.
   memset(&ble_cfg, 0x00, sizeof(ble_cfg));
   ble_cfg.conn_cfg.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;
   ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = NRF_SDH_BLE_GATT_MAX_MTU_SIZE;

   err_code = sd_ble_cfg_set(BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure number of custom UUIDS.
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.common_cfg.vs_uuid_cfg.vs_uuid_count = NRF_SDH_BLE_VS_UUID_COUNT;

   err_code = sd_ble_cfg_set(BLE_COMMON_CFG_VS_UUID, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the GATTS attribute table.
   memset(&ble_cfg, 0x00, sizeof(ble_cfg));
   ble_cfg.gatts_cfg.attr_tab_size.attr_tab_size = NRF_SDH_BLE_GATTS_ATTR_TAB_SIZE;

   err_code = sd_ble_cfg_set(BLE_GATTS_CFG_ATTR_TAB_SIZE, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure Service Changed characteristic.
   memset(&ble_cfg, 0x00, sizeof(ble_cfg));
   ble_cfg.gatts_cfg.service_changed.service_changed = NRF_SDH_BLE_SERVICE_CHANGED;

   err_code = sd_ble_cfg_set(BLE_GATTS_CFG_SERVICE_CHANGED, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure Write queue size write_cmd_tx_queue_size
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.conn_cfg.conn_cfg_tag = APP_BLE_CONN_CFG_TAG;
   ble_cfg.conn_cfg.params.gattc_conn_cfg.write_cmd_tx_queue_size = config->write_queue_size;

   err_code = sd_ble_cfg_set(BLE_CONN_CFG_GATTC, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Enable BLE stack.
   err_code = fnBleEnable(&ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure extended event length
   memset(&opt, 0x00, sizeof(opt));
   opt.common_opt.conn_evt_ext.enable = gpstBleParams->default_is_extended_evt;

   err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_EVT_EXT, &opt);
   APP_ERROR_CHECK(err_code);
}

__STATIC_INLINE void fnInitGapParams()
{
   uint32_t              err_code;
   ble_gap_conn_params_t   gap_conn_params;
   ble_gap_conn_sec_mode_t sec_mode;

   BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&sec_mode);

   memset(&gap_conn_params, 0, sizeof(gap_conn_params));

   gap_conn_params.min_conn_interval = gpstBleParams->default_conn_interval;
   gap_conn_params.max_conn_interval = gpstBleParams->default_conn_interval;

   gap_conn_params.slave_latency = gpstBleParams->default_slave_latency;
   gap_conn_params.conn_sup_timeout = gpstBleParams->default_conn_sup_timeout;

   err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
   APP_ERROR_CHECK(err_code);
}

__STATIC_INLINE uint32_t fnBleEnable(uint32_t * const p_app_ram_start)
{
   // Start of RAM, obtained from linker symbol.
   uint32_t const app_ram_start_link = *p_app_ram_start;

   uint32_t ret_code = sd_ble_enable(p_app_ram_start);
   if (*p_app_ram_start > app_ram_start_link)
   {
      LOG_DEBUG("Insufficient RAM allocated for the SoftDevice.\r\n");
      LOG_DEBUG("Change the RAM start location from 0x%x to 0x%x.\r\n", app_ram_start_link, *p_app_ram_start);
   }
   else
   {
      LOG_DEBUG("RAM starts at 0x%x\r\n", app_ram_start_link);
      if (*p_app_ram_start != app_ram_start_link)
      {
         LOG_DEBUG("RAM start location can be adjusted to 0x%x.\r\n", *p_app_ram_start);
      }
   }

   return ret_code;
}

__STATIC_INLINE void fnProcessEvent(const ble_evt_t * const p_ble_evt)
{
   uint16_t id = p_ble_evt->header.evt_id;
   uint16_t conn_handle = p_ble_evt->evt.common_evt.conn_handle;

   if ((BLE_CONN_HANDLE_INVALID != ghConnection) && (ghConnection != conn_handle))
   {
      LOG_DEBUG("Invalid connection handle 0x%x expected 0x%x\r\n", conn_handle, ghConnection);
      return;
   }

   ble_stack_evt_handler_t intHandler = fnGetEventHandler(id, 0); // default internal handler - level 0
   ble_stack_evt_handler_t extHandler = fnGetEventHandler(id, 1); // external handlers - TODO:: only 1 at the moment   
   
   if ((NULL != intHandler) && (NULL != extHandler))
   {
      /* execute internal event handler */
      intHandler(conn_handle, p_ble_evt);
      
      /* execulte external event handler */
      extHandler(conn_handle, p_ble_evt);
   }
   else if ((NULL != intHandler) && (NULL == extHandler))
   {
      /* execute internal event handler */
      intHandler(conn_handle, p_ble_evt);
      
      //LOG_DEBUG("Unexpected id (0x%x) internal handler not found...\r\n", id);
   }
   else if ((NULL == intHandler) && (NULL != extHandler))
   {
      /* execute external event handler */
      extHandler(conn_handle, p_ble_evt);
      
      //LOG_DEBUG("Unexpected id (0x%x) internal handler not found...\r\n", id);
   }
   else
   {
       LOG_DEBUG("Unexpected id (0x%x) handler not found...\r\n", id);
   }
}

__STATIC_INLINE ble_stack_evt_handler_t fnGetEventHandler(uint16_t id, uint32_t level)
{
   if (IS_GAP_EVT(id))
      return gfpGapEventHandlers[(id - BLE_GAP_EVT_BASE)][level];

   if (IS_GATTC_EVT(id))
      return gfpGattcEventHandlers[(id - BLE_GATTC_EVT_BASE)][level];
   
   return NULL;
}

__STATIC_INLINE bool fnSetEventHandler(const uint16_t id, const uint32_t level, const ble_stack_evt_handler_t handler)
{
   if (IS_GAP_EVT(id))
   {
      gfpGapEventHandlers[(id - BLE_GAP_EVT_BASE)][level] = handler;
      return true;
   }
   
   if (IS_GATTC_EVT(id))
   {
      gfpGattcEventHandlers[(id - BLE_GATTC_EVT_BASE)][level] = handler;
      return true;
   }
   
   return false;
}

__STATIC_INLINE void fnOnConnected(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ghConnection = conn_handle;
   LOG_DEBUG("Connection established 0x%x\r\n", ghConnection);

   ble_gap_conn_params_t connection_params = p_ble_evt->evt.gap_evt.params.connected.conn_params;
   ble_gap_conn_params_t desired_conn_params;

   sd_ble_gap_ppcp_get(&desired_conn_params);

   LOG_DEBUG("Client Connection Params:\r\n");
   LOG_DEBUG("min_conn_interval %u\r\n", connection_params.min_conn_interval);
   LOG_DEBUG("max_conn_interval %u\r\n", connection_params.max_conn_interval);
   LOG_DEBUG("conn_sup_timeout %u\r\n", connection_params.conn_sup_timeout);
   LOG_DEBUG("slave_latency %u\r\n", connection_params.slave_latency);
   UNUSED(connection_params);
   
   LOG_DEBUG("Desired Connection Params:\r\n");
   LOG_DEBUG("min_conn_interval %u\r\n", desired_conn_params.min_conn_interval);
   LOG_DEBUG("max_conn_interval %u\r\n", desired_conn_params.max_conn_interval);
   LOG_DEBUG("conn_sup_timeout %u\r\n", desired_conn_params.conn_sup_timeout);
   LOG_DEBUG("slave_latency %u\r\n", desired_conn_params.slave_latency);

   uint32_t err_code = sd_ble_gap_conn_param_update(conn_handle, &desired_conn_params);
   APP_ERROR_CHECK(err_code);
   
   const uint16_t desiredDataLength = 251;
   err_code = fnBleStack_DataLengthUpdate(conn_handle, desiredDataLength);
   APP_ERROR_CHECK(err_code);
}

__STATIC_INLINE uint32_t fnBleStack_DataLengthUpdate(const uint16_t conn_handle, uint16_t data_length)
{
      LOG_DEBUG("Updating data length to %u on connection 0x%x.",
                  data_length, conn_handle);
      
          ble_gap_data_length_params_t const dlp =
    {
        .max_rx_octets  = data_length,
        .max_tx_octets  = data_length,
        .max_rx_time_us = BLE_GAP_DATA_LENGTH_AUTO,
        .max_tx_time_us = BLE_GAP_DATA_LENGTH_AUTO,
    };
    
    ble_gap_data_length_limitation_t dll = {0};
    
    return sd_ble_gap_data_length_update(conn_handle, &dlp, &dll);
}

__STATIC_INLINE void fnOnConnectionParamsUpdated(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gap_conn_params_t connection_params = p_ble_evt->evt.gap_evt.params.conn_param_update.conn_params;

   LOG_DEBUG("Effective Connection Params:\r\n");
   LOG_DEBUG("min_conn_interval %u\r\n", connection_params.min_conn_interval);
   LOG_DEBUG("max_conn_interval %u\r\n", connection_params.max_conn_interval);
   LOG_DEBUG("conn_sup_timeout %u\r\n", connection_params.conn_sup_timeout);
   LOG_DEBUG("slave_latency %u\r\n", connection_params.slave_latency);
   UNUSED(connection_params);
}

__STATIC_INLINE void fnOnDisconnected(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ghConnection = BLE_CONN_HANDLE_INVALID;

   LOG_DEBUG("\r\nDisconnected 0x%x\r\n", conn_handle);
}

__STATIC_INLINE void fnOnConnParamUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
//    ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
//    ble_gap_evt_conn_param_update_request_t conn_param_update_request = p_gap_evt->params.conn_param_update_request;
//    ble_gap_conn_params_t const * p_conn_param_decided = conn_param_update_request.conn_params;
    
    ble_gap_conn_params_t const * p_conn_param_decided = &p_ble_evt->evt.gap_evt.params.conn_param_update_request.conn_params;
    uint32_t err_code = sd_ble_gap_conn_param_update(conn_handle, p_conn_param_decided);
}

__STATIC_INLINE void fnOnSecurityParamsRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   uint32_t err_code = sd_ble_gap_sec_params_reply(conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
   APP_ERROR_CHECK(err_code);

   LOG_DEBUG("BLE_GAP_EVT_SEC_PARAMS_REQUEST 0x%x\r\n", conn_handle);
}

__STATIC_INLINE void fnOnDataLengthUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

   uint8_t data_length_desired = NRF_SDH_BLE_GAP_DATA_LENGTH;
   // The SoftDevice only supports symmetric RX/TX data length settings.
   uint8_t const data_length_requested = p_gap_evt->params.data_length_update_request.peer_params.max_tx_octets;

   LOG_DEBUG("Peer on connection 0x%x requested a data length of %u bytes.\r\n", conn_handle, data_length_requested);

   uint8_t const data_length_effective = MIN(data_length_desired, data_length_requested);

   LOG_DEBUG("Updating data length to %u on connection 0x%x.\r\n", data_length_effective, conn_handle);

   ble_gap_data_length_params_t const dlp =
   {
      .max_rx_octets = data_length_effective,
      .max_tx_octets = data_length_effective,
      .max_rx_time_us = BLE_GAP_DATA_LENGTH_AUTO,
      .max_tx_time_us = BLE_GAP_DATA_LENGTH_AUTO,
   };

   ble_gap_data_length_limitation_t dll = { 0 };

   uint32_t err_code = sd_ble_gap_data_length_update(conn_handle, &dlp, &dll);
   APP_ERROR_CHECK(err_code); //TODO:: ???
}

__STATIC_INLINE void fnOnDataLengthUpdate(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
   uint16_t data_length_effective = p_gap_evt->params.data_length_update.effective_params.max_tx_octets;

   LOG_DEBUG("Data length updated to %u on connection 0x%0x.\r\n", data_length_effective, conn_handle);
   LOG_DEBUG("max_rx_octets: %u\r\n", p_gap_evt->params.data_length_update.effective_params.max_rx_octets);
   LOG_DEBUG("max_tx_octets: %u\r\n", p_gap_evt->params.data_length_update.effective_params.max_tx_octets);
   LOG_DEBUG("max_rx_time: %u\r\n", p_gap_evt->params.data_length_update.effective_params.max_rx_time_us);
   LOG_DEBUG("max_tx_time: %u\r\n", p_gap_evt->params.data_length_update.effective_params.max_tx_time_us);
   UNUSED(data_length_effective);
}

__STATIC_INLINE void fnOnPhyUpdateRequest(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
   ble_gap_phys_t const peer_preferred_phys = p_gap_evt->params.phy_update_request.peer_preferred_phys;
   
   LOG_DEBUG("Peer on connection 0x%x requested a phy of rx_phys %u tx_phys %u.\r\n", conn_handle, peer_preferred_phys.rx_phys, peer_preferred_phys.tx_phys);

   uint32_t err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &peer_preferred_phys);
   APP_ERROR_CHECK(err_code);
}

__STATIC_INLINE void fnOnPhyUpdate(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

   LOG_DEBUG("Phy updated status %u on connection 0x%0x.\r\n", p_gap_evt->params.phy_update.status, conn_handle);
   LOG_DEBUG("rx_phy: %u\r\n", p_gap_evt->params.phy_update.rx_phy);
   LOG_DEBUG("tx_phy: %u\r\n", p_gap_evt->params.phy_update.tx_phy);

   UNUSED(p_gap_evt);
}

__STATIC_INLINE void fnOnTimeout(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   LOG_DEBUG("BLE_TIMEOUT 0x%x\r\n", conn_handle);

   // TODO:: could be advertising timeout ???
}

__STATIC_INLINE void fnOnWriteResponse(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
  ble_gattc_evt_write_rsp_t write_rsp = p_ble_evt->evt.gattc_evt.params.write_rsp;
  
  if(0x0000 == p_ble_evt->evt.gattc_evt.gatt_status)
  {
    LOG_DEBUG("SUCCESS with write op: %02X\r\n", write_rsp.write_op);
  }
  else
  {
    LOG_DEBUG("NO SUCCESS\r\n");
  }
}


void fnMtuRequest()
{   
   uint16_t client_rx_mtu = 247;
   uint32_t err_code = sd_ble_gattc_exchange_mtu_request(ghConnection,client_rx_mtu );
}

static bool gbMtuResponseReceived = false;
__STATIC_INLINE void fnMtuResponse(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{
   ble_gattc_evt_exchange_mtu_rsp_t MTU_Size = p_ble_evt->evt.gattc_evt.params.exchange_mtu_rsp;   
   LOG_DEBUG("SUCCESS MTU is Size: %u\r\n", MTU_Size.server_rx_mtu );
   gbMtuResponseReceived = true;
}

bool fnBleStack_IsMtuSizeEstablished(void)
{
  return gbMtuResponseReceived;
}