
#include "device_info.h"
#include "app_error.h"

#include "ble_abstraction.h"
#include "version.h"
#include "flash.h"

#define BLE_UUID_DEVICE_INFORMATION_SERVICE                       0x180A     /**< Device Information service UUID. */
#define BLE_UUID_MODEL_NUMBER_STRING_CHAR                         0x2A24     /**< Model Number String characteristic UUID. */
#define BLE_UUID_SERIAL_NUMBER_STRING_CHAR                        0x2A25     /**< Serial Number String characteristic UUID. */
#define BLE_UUID_FW_REV_STRING_CHAR                               0x2A26     /**< Firmware Revision String characteristic UUID. */
#define BLE_UUID_HW_REV_STRING_CHAR                               0x2A27     /**< Hardware Revision String characteristic UUID. */
#define BLE_UUID_SW_REV_STRING_CHAR                               0x2A28     /**< Software Revision String characteristic UUID. */
#define BLE_UUID_MANUFACTURER_NAME_STRING_CHAR                    0x2A29     /**< Manufacturer Name String characteristic UUID. */

static ad_data_t gstDefaultAdvert =
{
   BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE,
   "NOTSET",
   { 0x3b , 0x4f , 0x8f, 0xe8, 0x22, 0xa3, 0xb9 , 0x87 , 0xf2 , 0x4d , 0xac, 0xf9, 0x01, 0x00, 0x09 , 0x1a },
   "RXW"
};

static ad_data_t *gpstAdvert = NULL;

// Helpers
__STATIC_INLINE void fnAddManufacturerCharacteristic(const uint16_t service_handle);
__STATIC_INLINE void fnAddSerialNumberCharacteristic(const uint16_t service_handle);
__STATIC_INLINE void fnAddVersionNumberCharacteristics(const uint16_t service_handle);
__STATIC_INLINE void fnAddStringCharacteristic(const uint16_t service_handle, const uint16_t string_id, uint8_t * p_string, const uint16_t length);


///
const ad_data_t* fnDeviceInfo_GetAdvert(void)
{
    return &gstDefaultAdvert;
}

///
void fnDeviceInfo_InitializeService(void)
{
   uint32_t      err_code = 0;
   uint16_t service_handle = 0;
   ble_uuid_t    ble_uuid = { 0 };

   // Add Device Information Service
   ble_uuid.type = BLE_UUID_TYPE_BLE;
   ble_uuid.uuid = BLE_UUID_DEVICE_INFORMATION_SERVICE;

   err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &service_handle);
   APP_ERROR_CHECK(err_code);

   fnAddManufacturerCharacteristic(service_handle);
   fnAddSerialNumberCharacteristic(service_handle);
   fnAddVersionNumberCharacteristics(service_handle);
}

///
bool fnDeviceInfo_SetDeviceName(char const* pName)
{
   const ad_data_t local = *fnDeviceInfo_GetAdvert();
   strcpy((char*)local.name, pName);

   flash_write_operation_t *op = fnFlash_BeginWrite(MSD_FLASH_START_ADDR, (uint8_t const*)&local, sizeof(ad_data_t));
   fnFlash_UpdateWriteChecksum(op);
   return fnFlash_EndWrite(op);
}

///
bool fnDeviceInfo_SetServiceName(char const* pName)
{
   const ad_data_t local = *fnDeviceInfo_GetAdvert();
   strcpy((char*)&local.service_name, pName);

   flash_write_operation_t *op = fnFlash_BeginWrite(MSD_FLASH_START_ADDR, (uint8_t const*)&local, sizeof(ad_data_t));
   fnFlash_UpdateWriteChecksum(op);
   return fnFlash_EndWrite(op);
}

__STATIC_INLINE void fnAddManufacturerCharacteristic(const uint16_t service_handle)
{
   uint8_t manufacturer[] = MANUFACTURER;
   fnAddStringCharacteristic(service_handle, BLE_UUID_MANUFACTURER_NAME_STRING_CHAR, manufacturer, sizeof(manufacturer));
}

__STATIC_INLINE void fnAddSerialNumberCharacteristic(const uint16_t service_handle)
{
   const ad_data_t *pAdvert = fnDeviceInfo_GetAdvert();
   const uint16_t length = fnAdData_GetNameLength(pAdvert) + 1; // +1 for null-terminated
   fnAddStringCharacteristic(service_handle, BLE_UUID_SERIAL_NUMBER_STRING_CHAR, (uint8_t*)pAdvert->name, length);
}

__STATIC_INLINE void fnAddVersionNumberCharacteristics(const uint16_t service_handle)
{
   uint8_t fwVersion[] = FW_VERSION;
   fnAddStringCharacteristic(service_handle, BLE_UUID_FW_REV_STRING_CHAR, fwVersion, sizeof(fwVersion));
}

__STATIC_INLINE void fnAddStringCharacteristic(const uint16_t service_handle, const uint16_t string_id, uint8_t * p_string, const uint16_t length)
{
   ble_gatts_char_handles_t char_handle = { 0 }; // don't think we need to keep these for read only chars

   ble_gatt_char_props_t char_props = { .read = 1 };

   ble_uuid_t          char_uuid = { 0 };
   ble_gatts_char_md_t char_md;
   ble_gatts_attr_t    attr_char_value;
   ble_gatts_attr_md_t attr_md;

   char_uuid.type = BLE_UUID_TYPE_BLE;
   char_uuid.uuid = string_id;

   memset(&attr_md, 0, sizeof(ble_gatts_attr_md_t));

   BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
   BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);

   attr_md.vloc = BLE_GATTS_VLOC_STACK;

   memset(&char_md, 0, sizeof(ble_gatts_char_md_t));
   char_md.char_props = char_props;

   memset(&attr_char_value, 0, sizeof(ble_gatts_attr_t));
   attr_char_value.p_uuid = &char_uuid;
   attr_char_value.p_attr_md = &attr_md;
   attr_char_value.max_len = length;

   attr_char_value.init_len = length;
   attr_char_value.p_value = p_string;

   uint32_t err_code = sd_ble_gatts_characteristic_add(service_handle,
      &char_md,
      &attr_char_value,
      &char_handle);

   APP_ERROR_CHECK(err_code);

   LOG_DEBUG("Attribute 0x%x value handle %u\r\n", string_id, char_handle.value_handle);
}

