
#include "dfu.h"
#include "crc32.h"
#include "bootloader_settings.h"
#include "flash.h"
#include "ble_abstraction.h"
#include "app_error.h"
#include "device_info.h"

#define BLE_UUID_ENTER_DFU_CHARACTERISTIC    (uint16_t)0x0004
#define DFU_SETTINGS_INIT_COMMAND_OFFSET     offsetof(nrf_dfu_settings_t, init_command)

#define REN_BLE_DFU_PIN_CODE                 "1984"

const uint8_t m_dfu_settings_buffer[NRF52_PAGE_SIZE]  __attribute__((section(".bootloader_settings_page")))
                                                      __attribute__((used));

static const uint8_t dfu_service_uuid[] = { 0x62, 0xD8, 0x7F, 0x4A, 0xC0, 0xBA, 0x4C, 0xB7, 0x8D, 0x2B, 0xA0, 0x65, 0x01, 0x00, 0x06, 0x66 };
static nrf_dfu_settings_t *gpBootloaderSettings = (nrf_dfu_settings_t *)&m_dfu_settings_buffer[0];
static ble_gatts_char_handles_t gstEnterDfuHandles = { 0 };

/* Helpers */
__STATIC_INLINE void fnAddEnterDfuCharacteristic(const ble_uuid_t char_uuid, const uint16_t service_handle, ble_gatts_char_handles_t *p_char_handle);
__STATIC_INLINE uint32_t fnCalculateCrc(nrf_dfu_settings_t const * p_settings);
__STATIC_INLINE uint32_t fnCalculateCrcForAdvName(nrf_dfu_adv_name_t const * p_settings);

void fnDfu_InitializeService(void)
{
   ble_uuid_t    ble_uuid = { 0 };
   uint16_t service_handle = 0;
   
   //fnBleStack_Subscribe(BLE_GATTS_EVT_WRITE, fnOnWrite);

   uint32_t err_code = sd_ble_uuid_vs_add((const ble_uuid128_t *)dfu_service_uuid, &ble_uuid.type);
   APP_ERROR_CHECK(err_code);

   // Get bytes 12-13 service uuid
   memcpy(&ble_uuid.uuid, &dfu_service_uuid[12], sizeof(uint16_t));

   // Add the service.
   err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &service_handle);
   APP_ERROR_CHECK(err_code);

   const ble_uuid_t dfu_char_uuid = { .uuid = BLE_UUID_ENTER_DFU_CHARACTERISTIC,.type = ble_uuid.type };
   fnAddEnterDfuCharacteristic(dfu_char_uuid, service_handle, &gstEnterDfuHandles);

   LOG_DEBUG("Enter DFU value handle %u\r\n", gstEnterDfuHandles.value_handle);

   LOG_DEBUG("Enter DFU setting %u\r\n", gpBootloaderSettings->enter_buttonless_dfu);
   UNUSED(gpBootloaderSettings);
}

/* TODO:: should we have just extended the ble_stack event subscriptions... */
void fnDfu_OnGattsEvtWrite(const uint16_t conn_handle, const ble_gatts_evt_write_t * const p_evt_write)
{
   if ((p_evt_write->handle == gstEnterDfuHandles.value_handle) && (p_evt_write->len == 4))
   {
      LOG_DEBUG("Enter DFU characteristic write %s len %u\r\n", p_evt_write->data, p_evt_write->len);

      if (strcmp(REN_BLE_DFU_PIN_CODE, (const char*)p_evt_write->data) == 0)
      {
         nrf_dfu_settings_t local = *gpBootloaderSettings;
         local.enter_buttonless_dfu = 1;
         local.crc = fnCalculateCrc(&local);

         // set the dfu advert name for the bootloader to use
         const ad_data_t * p_adv = fnDeviceInfo_GetAdvert();

         strcpy((char*)local.adv_name.name, (const char*)p_adv->name);
         strcat((char*)local.adv_name.name, "-DFU");
         local.adv_name.len = strlen((const char*)local.adv_name.name);
         local.adv_name.crc = fnCalculateCrcForAdvName(&local.adv_name);

         fnFlash_Write((uint32_t)m_dfu_settings_buffer, (uint8_t*)&local, sizeof(nrf_dfu_settings_t));
      }
   }
}

__STATIC_INLINE void fnAddEnterDfuCharacteristic(const ble_uuid_t char_uuid, const uint16_t service_handle, ble_gatts_char_handles_t *p_char_handle)
{
   ble_gatt_char_props_t char_props = { .write = 1,.write_wo_resp = 1 };

   ble_gatts_char_md_t char_md;
   ble_gatts_attr_t    attr_char_value;
   ble_gatts_attr_md_t attr_md;

   memset(&attr_md, 0, sizeof(ble_gatts_attr_md_t));

   BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.read_perm);
   BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

   attr_md.rd_auth = 0;
   attr_md.wr_auth = 0;
   attr_md.vlen = 1;
   attr_md.vloc = BLE_GATTS_VLOC_STACK;

   memset(&char_md, 0, sizeof(ble_gatts_char_md_t));

   char_md.char_props = char_props;

   static char user_desc[] = "Renishaw Enter DFU";

   char_md.p_char_user_desc = (uint8_t *)user_desc;
   char_md.char_user_desc_size = strlen(user_desc);
   char_md.char_user_desc_max_size = strlen(user_desc);

   memset(&attr_char_value, 0, sizeof(ble_gatts_attr_t));
   attr_char_value.p_uuid = &char_uuid;
   attr_char_value.p_attr_md = &attr_md;
   attr_char_value.max_len = 4;

   uint32_t err_code = sd_ble_gatts_characteristic_add(service_handle,
      &char_md,
      &attr_char_value,
      p_char_handle);

   APP_ERROR_CHECK(err_code);
}

__STATIC_INLINE uint32_t fnCalculateCrc(nrf_dfu_settings_t const * p_settings)
{
   // The crc is calculated from the s_dfu_settings struct, except the crc itself and the init command
   return crc32_compute((uint8_t*)(p_settings)+4, DFU_SETTINGS_INIT_COMMAND_OFFSET - 4, NULL);
}

__STATIC_INLINE uint32_t fnCalculateCrcForAdvName(nrf_dfu_adv_name_t const * p_adv_name)
{
   // Calculate the CRC for the structure excluding the CRC value itself.
   return crc32_compute((uint8_t*)p_adv_name + 4, sizeof(nrf_dfu_adv_name_t) - 4, NULL);
}
