#include "flash.h"

#include "app_error.h"

#include "sd_abstraction.h"
#include "timer.h"
#include "crc32.h"

#define MSD_FLASH_CONTENT_SIZE   1024 /* 1K storage to start */

__PACKED_STRUCT flash_write_operation_s
{
   bool is_active;
   uint32_t page;
   uint32_t page_address;
   uint32_t write_length;
   uint32_t *write_buffer;
};

__STATIC_INLINE bool fnWaitForFlashOperation(void);
__STATIC_INLINE bool fnIsFlashEvent(uint32_t evt_id);

/* manufacturer_specific_data section defined by linker script */
const uint8_t gabyMSDFlashPage[NRF52_PAGE_SIZE]    __attribute__((section(".manufacturer_specific_data")))
                                                   __attribute__((used));

static uint8_t gabyFlashPageBuffer[MSD_FLASH_CONTENT_SIZE] = { 0 };
static flash_write_operation_t gstAsyncWrite = { 0 }; /* only one outstanding at any time */

///
bool fnFlash_IsInitialized(void)
{
   uint32_t crc = *(uint32_t*)&gabyMSDFlashPage[FLASH_CRC32_OFFSET];
   uint32_t calc_crc = crc32_compute(gabyMSDFlashPage, FLASH_CRC32_OFFSET, NULL);

   return (crc == calc_crc) ? true : false; 
}

///
flash_write_operation_t * fnFlash_BeginWrite(const uint32_t addr, uint8_t const * data, const uint32_t length)
{
   const uint32_t page = addr / NRF52_PAGE_SIZE;
   uint32_t page_address = page * NRF52_PAGE_SIZE; /* address of start of page */
   const uint8_t  offset = addr - page_address;    /* offset into page to write new data  */

   APP_ASSERT(gstAsyncWrite.is_active == false);

   /* copy flash contents to write buffer */
   memcpy(gabyFlashPageBuffer, (uint8_t*)page_address, sizeof(gabyFlashPageBuffer));
   /* overwrite in write buffer */
   memcpy((uint8_t*)&gabyFlashPageBuffer[offset], data, length);
   
   gstAsyncWrite.page = page;
   gstAsyncWrite.page_address = page_address;
   gstAsyncWrite.write_length = (uint32_t)(sizeof(gabyFlashPageBuffer) / sizeof(uint32_t));
   gstAsyncWrite.write_buffer = (uint32_t*)gabyFlashPageBuffer;
   gstAsyncWrite.is_active = true;

   return &gstAsyncWrite;
}

///
void fnFlash_UpdateWriteChecksum(flash_write_operation_t* const op)
{
   APP_ASSERT(op->is_active == true);

   /* overwrite in write buffer - indexed in bytes... */
   uint8_t *p_bytes = (uint8_t*)op->write_buffer;

   uint32_t calc_crc = crc32_compute(p_bytes, FLASH_CRC32_OFFSET, NULL);
   *(uint32_t*)&p_bytes[FLASH_CRC32_OFFSET] = calc_crc;
}

///
bool fnFlash_EndWrite(flash_write_operation_t* const op)
{
   uint32_t err_code;

   APP_ASSERT(op->is_active == true);
   op->is_active = false;

   /* erase flash contents */
   err_code = sd_flash_page_erase(op->page);

   if (NRF_SUCCESS != err_code)
   {
      LOG_DEBUG("Failed to erase flash page 0x%x error %u\r\n", op->page, err_code);
      return false;
   }

   if (!fnWaitForFlashOperation())
      return false; // TODO::

   /* write buffer to flash page */
   err_code = sd_flash_write((uint32_t*)op->page_address, op->write_buffer, op->write_length);

   if (NRF_SUCCESS != err_code)
   {
      LOG_DEBUG("Failed to write flash %u\r\n", err_code);
      return false;
   }

   return fnWaitForFlashOperation();
}

///
bool fnFlash_Write(const uint32_t addr, uint8_t const * data, const uint32_t length)
{
   flash_write_operation_t *op = fnFlash_BeginWrite(addr, data, length);
   return fnFlash_EndWrite(op);
}

__STATIC_INLINE bool fnWaitForFlashOperation(void)
{
   uint32_t err_code;

   bool bBusy = true;
   bool bSuccess = false;

   handle_t hTimer = fnTimer_Set(1000, (bool)TIMER_SINGLESHOT);

   do
   {
      err_code = sd_app_evt_wait();
      while (1)
      {
         uint32_t evt_id;
         err_code = sd_evt_get(&evt_id);

         if (NRF_SUCCESS != err_code) // all events in queue have been processed, so break out of inner loop
         {
            break;
         }

         LOG_DEBUG("Event while waiting for flash %u\r\n", evt_id);

         if (fnIsFlashEvent(evt_id))
         {
            // only process one flash event
            bBusy = false;
            bSuccess = (NRF_EVT_FLASH_OPERATION_SUCCESS == evt_id);
         }
      }

   } while (bBusy && !fnTimer_HasExpired(hTimer));

   fnTimer_Kill(&hTimer);
   return bSuccess;
}

__STATIC_INLINE bool fnIsFlashEvent(uint32_t evt_id)
{
   return ((NRF_EVT_FLASH_OPERATION_ERROR == evt_id) ||
      (NRF_EVT_FLASH_OPERATION_SUCCESS == evt_id));
}