#include "stdtypes.h"
#include "hw.h"

#include "init_sys.h"
#include "soft_device.h"
#include "uart.h"
#include "rtps.h"
#include "ble_stack.h"
#include "gpio.h"
#include "timer.h"
#include "clock.h"

#define STARTUP_DELAY_US MSEC_TO_UNITS(100, UNIT_US)

static handle_t gh5msBleSendTimer = K_INVALID_TIMER;

__STATIC_INLINE void fnRtpsDataHandler(uint8_t const * p_data, uint16_t length);
__STATIC_INLINE void fnRtpsDataOutHandler();

typedef enum {
Init = 0,
Scan =1,
Connect =2,
Stream =3      
}State;

static handle_t ghBleSendTimeout = K_INVALID_TIMER;
static const uint32_t gkun10msTimerValue = 10;
static const uint32_t gkun5msTimerValue = 5;

///
int main(void)
{
#if DEBUG
   NVIC_SetPriority(DebugMonitor_IRQn, 6ul);
#endif

   LOG_DEBUG("Application Started\r\n"); 

   fnClock_Initialise(NRF_CLOCK, CoreDebug, DWT);
   fnClock_StartLF();  
   
   fnTimer_Initialise(NRF_RTC1);
   fnInitSys_Initialize(NRF_GPIO_PORT);   
   fnClock_Delay_us(STARTUP_DELAY_US);    /* 100ms startup delay following BT_POWER */
   fnUart_Inititalize(NRF_UART_PORT);
   
   fnSoftDevice_Enable();
   gh5msBleSendTimer = fnTimer_Set(gkun5msTimerValue, (bool)TIMER_REPEAT);
   
   fnRtps_Initialize(fnRtpsDataHandler); /* */
   fnTimer_Enable();
   ghBleSendTimeout = fnTimer_Set(gkun10msTimerValue, (bool)TIMER_SINGLESHOT);

   LOG_DEBUG("Application Running\r\n");
   
   State Current_State= Init;
   for (;;)
   {    
   switch(Current_State)
      {
         case Init:
            fnUart_Reset();
            Current_State = Scan;
            break;
         case Scan:          
            if(NRF_SUCCESS == fnBleStack_ScanStart())
            {
               Current_State = Connect;
            }
            break;
         case Connect:
            fnUart_RxPoll();
            fnUart_TxPoll();
            
            fnBleStack_PollEvents();
            
            if (!fnRtps_IsConnected())
            {
             if(fnBleStack_IsMtuSizeEstablished())
               fnRtps_EnableNotifications();
              else
              {
               /*Commands to do on initial connection.*/
               fnMtuRequest(); 
              }
            }
            else
            {
               Current_State = Stream;
            }
            
            break;
         case (Stream):
              if(fnRtps_IsConnected())
              {
                if(fnTimer_HasExpired(gh5msBleSendTimer))
                {
                   fnRtpsDataOutHandler();
                }
                 
                 fnUart_RxPoll();
                 fnUart_TxPoll();
                 fnBleStack_PollEvents();
              }
              else
              {
                  Current_State = Init;
              }
                     
            break;         
            
      }
            //fnSoftDevice_WaitOne(); /* wait for next event (asleep) */
   }
}
   
__STATIC_INLINE void fnRtpsDataHandler(uint8_t const * p_data, uint16_t length)
{
  /* This handler is used on receipt of a notification from the module. */
  fnUart_Write(p_data, length);
}

__STATIC_INLINE void fnRtpsDataOutHandler()
{
  /* This handler manages the writing to the module. */
   uint8_t mtu_packet[NRF_SDH_BLE_GATT_MAX_MTU_SIZE] = { 0 };
   uint32_t count = 0u;
   bool timedOut = false;
   bool buffering = true;
    
    timedOut = fnTimer_HasExpired(ghBleSendTimeout);
    buffering = (fnUart_GetReceivedDataLength() < 247);
    
    if (!buffering || timedOut)
    {
        count = fnUart_Read(mtu_packet, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
        bool success;
        if(count > 0)
            success = fnRtps_Write(mtu_packet, count);

        UNUSED(success);

        fnTimer_Kill(&ghBleSendTimeout);
        ghBleSendTimeout = fnTimer_Set(gkun10msTimerValue, (bool)TIMER_SINGLESHOT);
        timedOut = false;
    }
}
