
#include "rtps.h"

#include "ble_abstraction.h"
#include "ble_stack.h"
#include "app_error.h"
#include "device_info.h"
#include "dfu.h"
#include "uart.h"

//TODO:: what if these clash with the initialised service uuid (from eeprom)?
#define BLE_UUID_RTPS_MODULE_UUID            (uint16_t)0x0001     
#define BLE_UUID_RTPS_OUT_CHARACTERISTIC     (uint16_t)0x0002                                                                                       /**< The UUID of the RX Characteristic. */
#define BLE_UUID_RTPS_IN_CHARACTERISTIC      (uint16_t)0x0003                                                                                       /**< The UUID of the TX Characteristic. */
#define BLE_UUID_CCCD_DESCRIPTOR             (uint16_t)0x2902
#define BLE_NOTIFY_QUEUE_SIZE                7                                                                                            /**< Packets of size MTU */

#define N_AD_TYPES      2   
#define UUID16_SIZE     2   /**< Size of 16 bit UUID. */
#define UUID32_SIZE     4   /**< Size of 32 bit UUID. */
#define UUID128_SIZE    16  /**< Size of 128 bit UUID. */

/* Helpers */
__STATIC_INLINE bool ble_advdata_uuid_find(uint8_t  const * p_encoded_data,uint16_t data_len,ble_uuid_t const * p_target_uuid);
__STATIC_INLINE uint16_t ble_advdata_search(uint8_t const * p_encoded_data,uint16_t data_len,uint16_t * p_offset,uint8_t ad_type);

/* GATTS Event handlers */
__STATIC_INLINE void fnOnCharacteristicDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);

/* GAP Event Handlers */
__STATIC_INLINE void fnOnAdvertReport(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt);
__STATIC_INLINE void fnOnConnect(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);
__STATIC_INLINE void fnOnDisconnect(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);

/* GATTC Event Handlers */
__STATIC_INLINE void fnOnNotificationReceived(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);
__STATIC_INLINE void fnOnPrimaryServiceDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);
//__STATIC_INLINE void fnOnRelationshipDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);
__STATIC_INLINE void fnOnDescriptorsDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);
__STATIC_INLINE void fnOnWriteComplete(const uint16_t conn_handle, ble_evt_t const * p_ble_evt);

static ble_gattc_char_t gstInChar;
static ble_gattc_char_t gstOutChar;

static ble_gattc_desc_t gstInCccd;

static volatile bool gbIsNotifyEnabled = false;
static ble_stack_config_t gstConfig = { .write_queue_size = BLE_NOTIFY_QUEUE_SIZE };

static ble_rtps_data_handler_t gfpDataHandler;
static uint16_t guWriteBuffersAvailable;

static ble_gattc_handle_range_t handle_range;

/* The RTPS UUID which we are expecting to see. */
static ble_uuid_t const rtps_uuid= 
{
 .uuid = BLE_UUID_RTPS_MODULE_UUID,
 .type = BLE_UUID_TYPE_VENDOR_BEGIN
};

///
void fnRtps_Initialize(const ble_rtps_data_handler_t fpDataHandler)
{
   uint32_t      err_code = 0;
   ble_uuid_t    ble_uuid = { 0 };
   
   gfpDataHandler = fpDataHandler;
   guWriteBuffersAvailable = gstConfig.write_queue_size;

   fnBleStack_Initialize(&gstConfig);

   fnBleStack_Subscribe(BLE_GAP_EVT_CONNECTED, fnOnConnect); 
   fnBleStack_Subscribe(BLE_GAP_EVT_DISCONNECTED, fnOnDisconnect);
   fnBleStack_Subscribe(BLE_GAP_EVT_ADV_REPORT, fnOnAdvertReport);
   
   fnBleStack_Subscribe(BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP, fnOnPrimaryServiceDiscover);
   fnBleStack_Subscribe(BLE_GATTC_EVT_CHAR_DISC_RSP, fnOnCharacteristicDiscover);
   fnBleStack_Subscribe(BLE_GATTC_EVT_DESC_DISC_RSP, fnOnDescriptorsDiscover);
   
   fnBleStack_Subscribe(BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE, fnOnWriteComplete);
   fnBleStack_Subscribe(BLE_GATTC_EVT_HVX, fnOnNotificationReceived);
   
   // Add a custom base UUID - (ignores bytes 12-13 service uuid)
   const ad_data_t* pAdvert = fnDeviceInfo_GetAdvert();
   err_code = sd_ble_uuid_vs_add((const ble_uuid128_t *)pAdvert->service_uuid, &ble_uuid.type);
   APP_ERROR_CHECK(err_code);
}

///
bool fnRtps_IsConnected(void)
{
   return fnBleStack_IsConnected() && gbIsNotifyEnabled;
}

///
bool fnRtps_CanWrite(void)
{
   if (!fnRtps_IsConnected())
      return false;

   return (guWriteBuffersAvailable > 0) ? true : false;
}

///
bool fnRtps_Write(const uint8_t * const p_data, uint16_t length)
{   
   ble_gattc_write_params_t write_params;  
  
  if (!fnRtps_CanWrite())
      return false; /* exception ?*/

   if (length == 0)
      return false; /* exception ?*/

   write_params.flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_CANCEL;
   write_params.handle = gstOutChar.handle_value;
   write_params.len = length; 
   write_params.offset = 0;
   write_params.p_value = p_data;
   write_params.write_op = BLE_GATT_OP_WRITE_CMD;
   
   if(NRF_SUCCESS == fnBleStack_GattcWrite(&write_params))
   
   --guWriteBuffersAvailable; // TODO:: does this need to be atomic ???
   return true;
}

void fnRtps_EnableNotifications()
{ 
  ble_gattc_write_params_t notify_write_params = {0};
  
  uint8_t notification_enable[2] = { 0x01, 0x00 };
  
  notify_write_params.flags = BLE_GATT_EXEC_WRITE_FLAG_PREPARED_CANCEL;
  notify_write_params.handle = gstInCccd.handle;
  notify_write_params.len = 2; 
  notify_write_params.offset = 0; 
  notify_write_params.p_value = &notification_enable[0]; 
  notify_write_params.write_op = BLE_GATT_OP_WRITE_REQ;
  
  uint32_t err_code = fnBleStack_GattcWrite(&notify_write_params);
  
  if(NRF_SUCCESS == err_code)
  {
    gbIsNotifyEnabled = true;
  }
}

/**
Connection establisted to allow discovery of primary services
*/
__STATIC_INLINE void fnOnConnect(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
   uint32_t err_code = fnBleStack_PrimaryServicesDiscover(rtps_uuid.uuid,&rtps_uuid);
}

__STATIC_INLINE void fnOnDisconnect(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
  ble_gattc_char_t emptyChar = { 0 };
  ble_gattc_desc_t emptyDesc = { 0 };
  
   gbIsNotifyEnabled = false;
   handle_range.start_handle = 0;
   handle_range.end_handle = 0;
   guWriteBuffersAvailable = gstConfig.write_queue_size;
   gstInChar = emptyChar;
   gstOutChar = emptyChar;
   gstInCccd = emptyDesc;

   //fnUart_Reset();
}

/**
Find  the primary services available. Should be RTPS found here.
*/
__STATIC_INLINE void fnOnPrimaryServiceDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
    ble_gattc_evt_prim_srvc_disc_rsp_t const * p_evt_prim_srvc = &p_ble_evt->evt.gattc_evt.params.prim_srvc_disc_rsp;
    handle_range = p_evt_prim_srvc->services->handle_range;
    sd_ble_gattc_characteristics_discover(conn_handle,&handle_range);
}

/**
Find the characteristics of the RTPS in and RTPS out 
*/
__STATIC_INLINE void fnOnCharacteristicDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
    ble_gattc_evt_char_disc_rsp_t const * p_evt_char_disc_rsp = &p_ble_evt->evt.gattc_evt.params.char_disc_rsp;
    static ble_gattc_handle_range_t handle_range_char;
    handle_range_char = handle_range;
    handle_range_char.start_handle = p_evt_char_disc_rsp->chars->handle_value;

    if (0x00 != p_evt_char_disc_rsp->count )
    {
      if(BLE_UUID_RTPS_OUT_CHARACTERISTIC ==  p_evt_char_disc_rsp->chars->uuid.uuid)
      {
         gstOutChar = p_evt_char_disc_rsp->chars[0];
      }
      else if(BLE_UUID_RTPS_IN_CHARACTERISTIC ==  p_evt_char_disc_rsp->chars->uuid.uuid)
      {
         gstInChar = p_evt_char_disc_rsp->chars[0];
      }
      else
      {
         LOG_DEBUG("Unknown characteristic UUID");
      }
       sd_ble_gattc_characteristics_discover(conn_handle,&handle_range_char);
    }
    else
    {
      /* Once we have discovered the characteristics we need, then move onto descriptors. */
      sd_ble_gattc_descriptors_discover(conn_handle, &handle_range);
    }
}

/**
*/
__STATIC_INLINE void fnOnDescriptorsDiscover(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
    ble_gattc_evt_desc_disc_rsp_t const * p_evt_desc_disc_rsp = &p_ble_evt->evt.gattc_evt.params.desc_disc_rsp;
    static ble_gattc_handle_range_t handle_range_desc;
    handle_range_desc = handle_range;
    handle_range_desc.start_handle = (p_evt_desc_disc_rsp->descs->handle) + p_evt_desc_disc_rsp->count;
        
    if(BLE_UUID_CCCD_DESCRIPTOR == p_evt_desc_disc_rsp->descs->uuid.uuid)
    {
         gstInCccd = p_evt_desc_disc_rsp->descs[0];
         LOG_DEBUG("Found CCCD\r\n");
    }
    else if ( 0 != p_evt_desc_disc_rsp->count)
    {
         sd_ble_gattc_descriptors_discover(conn_handle, &handle_range_desc);
         LOG_DEBUG("Found descriptor %04X\r\n", p_evt_desc_disc_rsp->descs->uuid.uuid);
    }
    else
    {
         LOG_DEBUG("Found All descriptors.");
    }
}

/**

*/
__STATIC_INLINE void fnOnNotificationReceived(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
    ble_gattc_evt_hvx_t const * p_evt_hvx = &p_ble_evt->evt.gattc_evt.params.hvx;
    if(p_evt_hvx->type == BLE_GATT_HVX_NOTIFICATION)
    {
        gfpDataHandler(p_evt_hvx->data, p_evt_hvx->len);//data sent to UART
    }
    else if (p_evt_hvx->type == BLE_GATT_HVX_INDICATION)
    {
        LOG_DEBUG("indication");
        uint32_t err = sd_ble_gattc_hv_confirm(conn_handle,p_evt_hvx->handle); 
        if(err != 0)
          LOG_DEBUG("ERR: %d\r\n:", err);    
        
        gfpDataHandler(p_evt_hvx->data, p_evt_hvx->len);//data sent to UART        
    }
    
    LOG_DEBUG("data bytes received over BLE: %d \r\n", p_evt_hvx->len);
    for (int i = 0; i < p_evt_hvx->len; i++)
    {
         LOG_DEBUG(" %02X", p_evt_hvx->data[i]);
    }
    LOG_DEBUG("\r\n");    
}

__STATIC_INLINE void fnOnWriteComplete(const uint16_t conn_handle, ble_evt_t const * p_ble_evt)
{
   ble_gattc_evt_write_cmd_tx_complete_t const * p_evt_complete = &p_ble_evt->evt.gattc_evt.params.write_cmd_tx_complete;
   
   // TODO:: does this need to be atomic ??
   guWriteBuffersAvailable += p_evt_complete->count; // increase available buffers for notifications by the number of completed
}

/**
*/
__STATIC_INLINE void fnOnAdvertReport(const uint16_t conn_handle, const ble_evt_t * const p_ble_evt)
{   
    uint32_t err_code;
    ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;
    ble_gap_evt_adv_report_t const * p_adv_report = &p_gap_evt->params.adv_report;
    
    if (ble_advdata_uuid_find(p_adv_report->data.p_data, p_adv_report->data.len, &rtps_uuid))
    {
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[0]);
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[1]);
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[2]);
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[3]);
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[4]);
         LOG_DEBUG(" %02X", p_adv_report->peer_addr.addr[5]);
        if (p_adv_report->peer_addr.addr[0] == 0xc0
         && p_adv_report->peer_addr.addr[1] == 0xd9
         && p_adv_report->peer_addr.addr[2] == 0xc8
         && p_adv_report->peer_addr.addr[3] == 0x3a
         && p_adv_report->peer_addr.addr[4] == 0xa3
         && p_adv_report->peer_addr.addr[5] == 0xd7)
        {
            err_code = fnBleStack_Connect(&p_adv_report->peer_addr);
        
            if (err_code == NRF_SUCCESS)
            {
                // scan is automatically stopped by the connect
                LOG_DEBUG("Connecting to target %02x%02x%02x%02x%02x%02x",
                     p_adv_report->peer_addr.addr[5],
                     p_adv_report->peer_addr.addr[4],
                     p_adv_report->peer_addr.addr[3],
                     p_adv_report->peer_addr.addr[2],
                     p_adv_report->peer_addr.addr[1],
                     p_adv_report->peer_addr.addr[0]
                     );            
            }
        }
    }
    else
    {
        err_code = fnBleStack_ScanResume();
        APP_ERROR_CHECK(err_code);
    }
}

__STATIC_INLINE bool ble_advdata_uuid_find(uint8_t  const * p_encoded_data,uint16_t data_len,ble_uuid_t const * p_target_uuid)
{
    uint32_t            err_code;
    uint16_t            data_offset = 0;
    uint8_t             raw_uuid_len = UUID128_SIZE;
    uint8_t const *     p_parsed_uuid;
    uint16_t            parsed_uuid_len = data_len;
    uint8_t             raw_uuid[UUID128_SIZE];
    uint8_t             ad_types[N_AD_TYPES];

 

    err_code = sd_ble_uuid_encode(p_target_uuid, &raw_uuid_len, raw_uuid);

    if ((p_encoded_data == NULL) || (err_code != NRF_SUCCESS))
    {
        // Invalid p_encoded_data or p_target_uuid.
        return false;
    }

    /* Replace Switch Case with this as this is what we are expecting. */
   ad_types[0] = BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_COMPLETE;
   ad_types[1] = BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_MORE_AVAILABLE;

    for (uint8_t i = 0; (i < N_AD_TYPES) && (data_offset == 0); i++)
    {
        parsed_uuid_len = ble_advdata_search(p_encoded_data, data_len, &data_offset, ad_types[i]);
    }
    
    if (data_offset == 0)
    {
        // Could not find any relevant UUIDs in the encoded data.
        return false;
    }

    p_parsed_uuid = &p_encoded_data[data_offset];

    /* Print the full found UUID. */
    LOG_DEBUG("\r\nUUID = ");
    for(uint8_t i = 0; i< parsed_uuid_len; i ++)
    {
        LOG_DEBUG("%02X", p_parsed_uuid[i]);
    }
    LOG_DEBUG("\r\n");
    
    // Verify if any UUID matches the given UUID.
    for (uint16_t list_offset = 0; list_offset < parsed_uuid_len; list_offset += raw_uuid_len)
    {
        if (memcmp(&p_parsed_uuid[list_offset], raw_uuid, raw_uuid_len) == 0)
        {
            return true;
        }
    }
    
    // Could not find the UUID among the encoded data.
    return false;
}

__STATIC_INLINE uint16_t ble_advdata_search(uint8_t const * p_encoded_data,
                            uint16_t        data_len,
                            uint16_t      * p_offset,
                            uint8_t         ad_type)
{
    if ((p_encoded_data == NULL) || (p_offset == NULL))
    {
        return 0;
    }
    
    uint16_t i = 0;

    while (((i < *p_offset) || (p_encoded_data[i + 1] != ad_type)) && (i < data_len))
    {
        // Jump to next data.
        i += (p_encoded_data[i] + 1);
    }

    if (i >= data_len)
    {
        return 0;
    }
    else
    {
        *p_offset = i + 2;
        return (p_encoded_data[i] - 1);
    }
}


