
#include "soft_device.h"

#include "sd_abstraction.h"
#include "clock.h"

#include "app_error.h"

//
#define CRITICAL_REGION_ENTER()                                                             \
    {                                                                                       \
        uint8_t __CR_NESTED = 0;                                                            \
        sd_nvic_critical_region_enter(&__CR_NESTED);
#define CRITICAL_REGION_EXIT()                                                              \
        sd_nvic_critical_region_exit(__CR_NESTED);                                          \
    }

#if defined(__GNUC__)
#define NRF_BREAKPOINT
#define DISABLE_IRQ
#else
#define DISABLE_IRQ __disable_irq()
#define NRF_BREAKPOINT __BKPT(0)
#endif

#define NRF_BREAKPOINT_COND do {                            \
   /* C_DEBUGEN == 1 -> Debugger Connected */               \
   if (CoreDebug->DHCSR & CoreDebug_DHCSR_C_DEBUGEN_Msk)    \
   {                                                        \
      /* Generate breakpoint if debugger is conncted */     \
      NRF_BREAKPOINT;                                       \
   }                                                        \
} while (0)

/* Global nvic state instance, required by nrf_nvic.h */
nrf_nvic_state_t nrf_nvic_state;

/* Helpers */
__STATIC_INLINE uint32_t enable_softdevice(void);

void fnSoftDevice_Enable(void)
{
   uint32_t err_code = enable_softdevice();
   APP_ERROR_CHECK(err_code);

   LOG_DEBUG("Soft device enabled\r\n");

   err_code = sd_power_dcdc_mode_set(NRF_POWER_DCDC_ENABLE);
   APP_ERROR_CHECK(err_code);
}

void fnSoftDevice_WaitOne(void)
{
   uint32_t err_code = sd_app_evt_wait(); 
   
   if (err_code != NRF_SUCCESS)
   {
      LOG_DEBUG("sd_app_evt_wait err_code %x\r\n", err_code);
   }
   else
   {
      //LOG_DEBUG(".");
   }
}

__STATIC_INLINE uint32_t enable_softdevice(void)
{
   uint32_t ret_code;

   //LF clock may already be started...

   CRITICAL_REGION_ENTER();
   ret_code = sd_softdevice_enable(&clock_lf_cfg, app_error_fault_handler);
   CRITICAL_REGION_EXIT();

   if (ret_code != NRF_SUCCESS)
   {
      return ret_code;
   }

   // Enable event interrupt.
   // Interrupt priority has already been set by the stack.
   ret_code = sd_nvic_EnableIRQ((IRQn_Type)SD_EVT_IRQn);
   return ret_code;
}

__WEAK void app_error_fault_handler(uint32_t id, uint32_t pc, uint32_t info)
{
   DISABLE_IRQ;

#ifndef DEBUG
   LOG_DEBUG("Fatal error");
#else
   switch (id)
   {
   case NRF_FAULT_ID_SD_ASSERT:
      LOG_DEBUG("SOFTDEVICE: ASSERTION FAILED");
      break;
   case NRF_FAULT_ID_APP_MEMACC:
      LOG_DEBUG("SOFTDEVICE: INVALID MEMORY ACCESS");
      break;
   default:
      LOG_DEBUG("UNKNOWN FAULT at 0x%08X", pc);
      break;
   }
#endif

   NRF_BREAKPOINT_COND;
   // On assert, the system can only recover with a reset.

#ifndef DEBUG
   LOG_DEBUG("System reset");
   NVIC_SystemReset();
#else

#endif // DEBUG
}

void SD_EVT_IRQHandler(void)
{
   /* implemented to clear pending interrupt */
   /* -> using Thread Mode Event retrieval */
}
