// Ported from 8051 version

#include "timer.h"

#include "gpio.h"
#include "hw.h"

typedef struct timer_entry_t timer_entry_t;

__PACKED_STRUCT timer_entry_t
{
   timer_mode_t eState;
   uint32_t uReload;
   uint32_t uCount;
};

#define K_TIMER_LIST_SIZE (8)

static NRF_RTC_Type * gpRtcRegister;
static timer_entry_t gastTimerList[K_TIMER_LIST_SIZE];

/* gbmTimerSignalled & gbmTimerActive are bit flags corresponding to timers
* in gastTimerList. gbmTimerSignalled indicates that timer has expired.
* gbmTimerActive indicates that timer is in use. */
volatile uint8_t gbmTimerSignalled;
volatile uint8_t gbmTimerActive;


/*!
* \brief
* Configures the timer for 5ms interrupts.
*
*/
void fnTimer_Initialise(NRF_RTC_Type * const pRtcRegister)
{
   gpRtcRegister = pRtcRegister;

   gpRtcRegister->TASKS_STOP = 0;  // not running

   gpRtcRegister->PRESCALER = 163; // 5ms this has to be coarse resolution due to soft device interrupt processing delays...

   gpRtcRegister->EVTENCLR = UINT32_MAX;
   gpRtcRegister->INTENCLR = UINT32_MAX;

   gpRtcRegister->EVTENSET = RTC_EVTENSET_TICK_Msk;
   gpRtcRegister->INTENSET = RTC_INTENSET_TICK_Msk;
   gpRtcRegister->TASKS_CLEAR = 1;

   NVIC_SetPriority(RTC1_IRQn, 7ul); //--lowest priority
   NVIC_ClearPendingIRQ(RTC1_IRQn);
   NVIC_EnableIRQ(RTC1_IRQn);

   gbmTimerActive = 0u;
   gbmTimerSignalled = 0u;
}


/*!
* \brief
* Enables Timer
*
*/
void fnTimer_Enable(void)
{
   gpRtcRegister->TASKS_START = 1;
}


/*!
* \brief
* Stops the timer.
*
*/
void fnTimer_Disable(void)
{
   gpRtcRegister->TASKS_STOP = 1;
}


/*!
* \brief
* Tests to see if the timer has expired
*
* \param hTimer
* is the handle (index) to the timer
*
* \returns
* TRUE - if the timer is active and has expired, FALSE - otherwise.
*
*
*/
bool fnTimer_HasExpired(handle_t hTimer)
{
   bool bRetVal;
   uint8_t bt;

   if ((hTimer >= K_TIMER_LIST_SIZE) || (hTimer < 0))
   {
      return false;
   }

   bt = (1 << hTimer);

   bRetVal = (gbmTimerActive & bt) && (gbmTimerSignalled & bt);  /*double checked locking*/

   if (bRetVal)
   {
      gpRtcRegister->INTENCLR = RTC_INTENCLR_TICK_Msk;    /*disable interrupt */

      bRetVal = (gbmTimerActive & bt) && (gbmTimerSignalled & bt);
      gbmTimerSignalled &= ~bt;

      gpRtcRegister->INTENSET = RTC_INTENSET_TICK_Msk;    /*enable interrupt */
   }

   return bRetVal;
}


/*!
* \brief
* Reloads the timer indicated by hTimer from the list
*
* \param hTimer
* is the handle (index) to the timer
*
* \param uPeriod
* is the timeout value in ms.
*
* \returns
* true - if the timer was successfully reloaded, false - otherwise.
*
* Write detailed description for fnTimer_Reload here.
*
* \remarks
* A timer may only be reloaded if it is a single-shot timer that has expired
* and has not yet been killed.
*
*/
bool fnTimer_Reload(handle_t hTimer, uint32_t uPeriod)
{
   uint8_t bt;
   bool bVal;

   if ((hTimer >= K_TIMER_LIST_SIZE) || (hTimer < 0))
   {
      return false;
   }

   bt = (1 << hTimer);

   if (!(gbmTimerActive & bt))
   {
      return false;
   }

   uint32_t uTick = MSEC_TO_UNITS(uPeriod, UNIT_5_MS);
   gbmTimerActive &= ~bt;  /*disable timer temporarily*/

   bVal = false;
   if (gastTimerList[hTimer].eState == TIMER_EXPIRED)
   {
      gastTimerList[hTimer].eState = TIMER_SINGLESHOT;
      gastTimerList[hTimer].uReload = uTick;
      gastTimerList[hTimer].uCount = uTick;
      gbmTimerSignalled &= ~bt;
      bVal = true;
   }

   gbmTimerActive |= bt;
   return bVal;
}

void fnTimer_Restart(handle_t hTimer)
{
   uint8_t bt;

   if ((hTimer >= K_TIMER_LIST_SIZE) || (hTimer < 0))
   {
      return;
   }

   bt = (1 << hTimer);

   if (!(gbmTimerActive & bt))
   {
      return;
   }

   gbmTimerActive &= ~bt;  /*disable timer temporarily*/

   gastTimerList[hTimer].uCount = gastTimerList[hTimer].uReload;
   gbmTimerSignalled &= ~bt;

   gbmTimerActive |= bt;
}

/*!
* \brief
* Removes the timer indicated by hTimer from the list
*
* \param phTimer
* is a pointer to the handle (index) of the timer
*
* Write detailed description for fnTimer_Kill here.
*
* \remarks
* finally phTimer points to K_INVALID_TIMER
*
*/
void fnTimer_Kill(handle_t *phTimer)
{
   uint8_t bt;

   if ((*phTimer >= K_TIMER_LIST_SIZE) || (*phTimer < 0))
      return;

   bt = (1 << *phTimer);
   gbmTimerActive &= ~bt;
   gbmTimerSignalled &= ~bt;
   *phTimer = K_INVALID_TIMER;
}



/*!
* \brief
* Clients call this function to request a timer.
*
* \param uPeriod
* is the timeout value in ms.
*
* \param bRepeat
* indicates whether the timer should be a repeat or single-shot.
*
* \returns
* A handle (index) to the timer or K_INVALID_TIMER if there are no timers available.
*
* Write detailed description for fnTimer_Set here.
*
*/
handle_t fnTimer_Set(uint32_t uPeriod, bool bRepeat)
{
   /*uPeriod == 0*/
   handle_t i;
   uint8_t bt = 1;

   uint32_t uTick = MSEC_TO_UNITS(uPeriod, UNIT_5_MS);
   for (i = 0; i < K_TIMER_LIST_SIZE; i++, bt <<= 1)
   {
      if (!(gbmTimerActive & bt))
      {
         gastTimerList[i].uReload = uTick;
         gastTimerList[i].uCount = uTick;
         gastTimerList[i].eState = (bRepeat) ? TIMER_REPEAT : TIMER_SINGLESHOT;
         gbmTimerSignalled &= ~bt;
         gbmTimerActive |= bt;
         return i;
      }
   }

   return K_INVALID_TIMER;
}

//static volatile bool bin = false; // TODO:: TEMP
/*!
* \brief
* Timer ISR
*
* This interrupt service routine services RTC 1. It scans the list of
* software timers. Whenever it finds an active timer it decrements its count.
* If that timer reaches zero then that time is signalled. After which
* the software timer is reloaded if it is a repeat timer.
*
* \remarks
* Note the client is responsible for killing the timer. *
*
*/

void RTC1_IRQHandler(void)
{
   handle_t i;
   uint8_t bt = 1u;

   //(bin) ? fnGPIO_ClearPin(NRF_GPIO_PORT, BT_STATUS_PIN_NUMBER) : fnGPIO_SetPin(NRF_GPIO_PORT, BT_STATUS_PIN_NUMBER);
   //bin = !bin;

   for (i = 0; i < K_TIMER_LIST_SIZE; i++, bt <<= 1)
   {
      if (gbmTimerActive & bt)
      {
         timer_entry_t* pstTimerEntry = gastTimerList + i;
         if (pstTimerEntry->uCount)
         {
            pstTimerEntry->uCount--;
            if (pstTimerEntry->uCount == 0)
            {
               gbmTimerSignalled |= bt;
               if (pstTimerEntry->eState == TIMER_REPEAT)
               {
                  pstTimerEntry->uCount = pstTimerEntry->uReload;
               }
               else
               {
                  pstTimerEntry->eState = TIMER_EXPIRED;
               }
            }
         }
      }
   }

   gpRtcRegister->TASKS_CLEAR = 1;

   // Clear all events (also unexpected ones)
   gpRtcRegister->EVENTS_COMPARE[0] = 0;
   gpRtcRegister->EVENTS_COMPARE[1] = 0;
   gpRtcRegister->EVENTS_COMPARE[2] = 0;
   gpRtcRegister->EVENTS_COMPARE[3] = 0;
   gpRtcRegister->EVENTS_TICK = 0;
   gpRtcRegister->EVENTS_OVRFLW = 0;
}

