
#include "uart.h"

#include "gpio.h"
#include "clock.h"

#define FIFO_SIZE (RX_BUFFER_SIZE) /*needs to be binary factor e.g. 256, 512, 1024*/
#define FIFO_WRAP (FIFO_SIZE - 1) 

static NRF_UART_Type * gpUartRegister;

// Receiver
static volatile uint32_t gunReceiverHead;
static volatile uint32_t gunReceiverTail;
static uint32_t gunCachedReceiverHead;
static uint32_t gunCachedReceiverTail;
// Transmitter
static volatile uint32_t gunTransmitHead;
static volatile uint32_t gunTransmitTail;
static uint32_t gunCachedTransmitHead;
static uint32_t gunCachedTransmitTail;

static volatile bool gbTxComplete; /* Indicates that transmission is complete*/

static uint8_t gaReceiveBuffer[FIFO_SIZE];
static uint8_t gaTransmitBuffer[FIFO_SIZE];

static volatile bool gbIsBreakDetected = false;
static volatile bool gbIsErrorDetected = false;

//Helpers
__STATIC_INLINE bool fnIsInterruptEnabled(NRF_UART_Type * p_reg, uint32_t int_mask);
__STATIC_INLINE bool fnIsReceiveEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE bool fnIsTransmitEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE bool fnIsErrorEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE void fnClearReceiveEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE void fnClearTransmitEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE void fnClearErrorEvent(NRF_UART_Type * p_reg);
__STATIC_INLINE void fnTxByte(void);
__STATIC_INLINE void fnHandleReceive(void);
__STATIC_INLINE void fnHandleTransmit(void);

///
void fnUart_Inititalize(NRF_UART_Type * const pUartRegister)
{
   gpUartRegister = pUartRegister;   
   
   gpUartRegister->BAUDRATE = UART_BAUDRATE_BAUDRATE_Baud115200;
   gpUartRegister->CONFIG = (uint32_t)(UART_CONFIG_PARITY_Excluded << UART_CONFIG_PARITY_Pos) | (uint32_t)UART_CONFIG_HWFC_Enabled;

   // Configure UART associated GPIO pins
   gpUartRegister->PSELRXD = BT_RX_PIN_NUMBER;
   gpUartRegister->PSELTXD = BT_TX_PIN_NUMBER;

   gpUartRegister->PSELRTS = BT_RTS_PIN_NUMBER;
   gpUartRegister->PSELCTS = BT_CTS_PIN_NUMBER;

   // manual flow control in code below
   //gpUartRegister->PSELRTS = 0xFFFFFFFF;
   //gpUartRegister->PSELCTS = 0xFFFFFFFF;

   gbTxComplete = true;  /*indicate that we are ready to transmit*/

   gunReceiverHead = 0;
   gunReceiverTail = 0;
   gunCachedReceiverHead = 0;
   gunCachedReceiverTail = 0;
   gunTransmitHead = 0;
   gunTransmitTail = 0;
   gunCachedTransmitHead = 0;
   gunCachedTransmitTail = 0;
   
   fnClearTransmitEvent(gpUartRegister);
   fnClearReceiveEvent(gpUartRegister);
   fnClearErrorEvent(gpUartRegister);

   gpUartRegister->INTENCLR = 0xFFFFFFFF;
   gpUartRegister->INTENSET = UART_INTENSET_TXDRDY_Msk | UART_INTENSET_RXDRDY_Msk | UART_INTENSET_ERROR_Msk;

   // enable interrupts
   NVIC_SetPriority(UARTE0_UART0_IRQn, 2ul);
   NVIC_ClearPendingIRQ(UARTE0_UART0_IRQn);
   NVIC_EnableIRQ(UARTE0_UART0_IRQn);

   gpUartRegister->ENABLE = UART_ENABLE_ENABLE_Enabled;
   gpUartRegister->TASKS_STARTRX = 0x01ul;

   gbIsBreakDetected = false;
}

///
void fnUart_RxPoll(void) /* RxPoll synchronizes cached head and tail pointers safely with interrupts disabled */
{
   gpUartRegister->INTENCLR = UART_INTENSET_RXDRDY_Msk;

   gunCachedReceiverHead = gunReceiverHead; /*sync cached pointers*/
   gunReceiverTail = gunCachedReceiverTail;

   uint32_t space = FIFO_SIZE - ((gunCachedReceiverHead - gunCachedReceiverTail) & FIFO_WRAP);
   if (space > 1UL) /*space cleared - carry on*/
   {
      fnGPIO_ClearPin(BT_RTS_PIN_NUMBER);
   }

   gpUartRegister->INTENSET = UART_INTENSET_RXDRDY_Msk;
}

///
void fnUart_TxPoll(void)
{
   if (gbTxComplete) /*there is no transmission in progress*/
   {
      gunTransmitHead = gunCachedTransmitHead; /*sync cached pointers*/
      gunCachedTransmitTail = gunTransmitTail;

      if (gunCachedTransmitHead != gunCachedTransmitTail) /*data to be transmitted*/
      {
         gbTxComplete = false; //transmit in progress

         fnClearTransmitEvent(gpUartRegister);
         gpUartRegister->TASKS_STARTTX = 0x01ul;

         fnTxByte();
      }
   }
}

///
void fnUart_Reset(void)
{  
  gpUartRegister->TASKS_SUSPEND = 0x01ul;
  
  gpUartRegister->ENABLE = UART_ENABLE_ENABLE_Disabled;
  
  NVIC_ClearPendingIRQ(UARTE0_UART0_IRQn);
  
  //Coppied from fnUart_SendBreak() should create a break to show when device cut 
  
   // First disconnect the transmit pin from the UART by 
   // writing -1 to the PSELTXD register
   gpUartRegister->PSELTXD = 0xFFFFFFFF;

   // Set pin low to generate the necessary zeros on the line
   fnGPIO_ClearPin(BT_TX_PIN_NUMBER);

   // Delay for enough time to generate a break
   fnClock_Delay_us(200);

   // Reconnect the pin to the UART
   gpUartRegister->PSELTXD = BT_TX_PIN_NUMBER;
  
  // clears errors? 
  fnUart_ClearError();
  
  //if there has been a uart timeout buffer error then ERRORSRC is set 
   gpUartRegister->ERRORSRC = 0x00ul;
   
   fnClearTransmitEvent(gpUartRegister);
   fnClearReceiveEvent(gpUartRegister);
   fnClearErrorEvent(gpUartRegister);
   
   gbIsBreakDetected = false;
   gbIsErrorDetected = false;

   gunReceiverHead = 0;
   gunReceiverTail = 0;
   gunCachedReceiverHead = 0;
   gunCachedReceiverTail = 0;
   gunTransmitHead = 0;
   gunTransmitTail = 0;
   gunCachedTransmitHead = 0;
   gunCachedTransmitTail = 0;
   
   uint8_t data = gpUartRegister->RXD;
   
   for (uint32_t i = 0; i < FIFO_SIZE; i++)
   {
      gaReceiveBuffer[i] = 0x00ul;
      gaTransmitBuffer[i] = 0x00ul;
   }
   
   gpUartRegister->INTENCLR = 0xFFFFFFFF;
   gpUartRegister->INTENSET = UART_INTENSET_TXDRDY_Msk | UART_INTENSET_RXDRDY_Msk | UART_INTENSET_ERROR_Msk;
   
   gpUartRegister->ENABLE = UART_ENABLE_ENABLE_Enabled;
   

   NVIC_EnableIRQ(UARTE0_UART0_IRQn);
   
   gpUartRegister->TASKS_STARTRX = 0x01ul;
   
   gbTxComplete = true;  /*indicate that we are ready to transmit*/
}

///
uint32_t fnUart_Read(uint8_t * const pBuffer, const uint32_t unCount)
{
   uint32_t length = (gunCachedReceiverHead - gunCachedReceiverTail) & FIFO_WRAP;
   uint32_t count = unCount;

   if (length < count) //truncate
      count = length;

   for (uint32_t i = 0; i < count; i++)
   {
      *(pBuffer + i) = gaReceiveBuffer[gunCachedReceiverTail];
      gunCachedReceiverTail++;
      gunCachedReceiverTail &= FIFO_WRAP;
   }
   
   return count;
}

uint32_t fnUart_GetReceivedDataLength(void)
{
    return (gunCachedReceiverHead - gunCachedReceiverTail) & FIFO_WRAP;
}

///
bool fnUart_Write(const uint8_t * pBuffer, uint32_t unCount)
{
    if (unCount >= FIFO_SIZE)
    {
        return false; // more of an exception (might add CException support later ???)
    }

   /*is there enough space for bytes?*/
   uint32_t unSize = (gunCachedTransmitTail - gunCachedTransmitHead) & FIFO_WRAP;
   if ((unSize != 0) && (unSize <= unCount))
   {
      return false; /*full*/
   }

   while (unCount--)
   {
      gaTransmitBuffer[gunCachedTransmitHead] = *pBuffer++;
      LOG_DEBUG(" sent: %x", gaTransmitBuffer[gunCachedTransmitHead]);
      gunCachedTransmitHead++;
      gunCachedTransmitHead &= FIFO_WRAP; /*circular buffer 512 bytes*/
   }

   LOG_DEBUG("\r\n");
   return true;
}

bool fnUart_IsBreak(void)
{
   return gbIsBreakDetected;
}

void fnUart_SendBreak(void)
{
   gpUartRegister->ENABLE = UART_ENABLE_ENABLE_Disabled;

   // First disconnect the transmit pin from the UART by 
   // writing -1 to the PSELTXD register
   gpUartRegister->PSELTXD = 0xFFFFFFFF;

   // Set pin low to generate the necessary zeros on the line
   fnGPIO_ClearPin(BT_TX_PIN_NUMBER);

   // Delay for enough time to generate a break
   fnClock_Delay_us(200);

   // Reconnect the pin to the UART
   gpUartRegister->PSELTXD = BT_TX_PIN_NUMBER;

   gpUartRegister->ENABLE = UART_ENABLE_ENABLE_Enabled;
}

void fnUart_ClearBreak(void)
{
   gbIsBreakDetected = false;
}

bool fnUart_IsError(void)
{ 
   return gbIsErrorDetected; 
}

void fnUart_ClearError(void)
{
   gbIsErrorDetected = false;
}

__STATIC_INLINE bool fnIsInterruptEnabled(NRF_UART_Type * p_reg, uint32_t int_mask)
{
   return (bool)(p_reg->INTENSET & int_mask);
}

__STATIC_INLINE bool fnIsReceiveEvent(NRF_UART_Type * p_reg)
{
   return (bool)(p_reg->EVENTS_RXDRDY);
}

__STATIC_INLINE bool fnIsTransmitEvent(NRF_UART_Type * p_reg)
{
   return (bool)(p_reg->EVENTS_TXDRDY);
}

__STATIC_INLINE bool fnIsErrorEvent(NRF_UART_Type * p_reg)
{
   return (bool)(p_reg->EVENTS_ERROR);
}

__STATIC_INLINE void fnClearReceiveEvent(NRF_UART_Type * p_reg)
{
   p_reg->EVENTS_RXDRDY = 0x0UL;
   uint32_t dummy = p_reg->EVENTS_RXDRDY;
   (void)dummy;
}

__STATIC_INLINE void fnClearTransmitEvent(NRF_UART_Type * p_reg)
{
   p_reg->EVENTS_TXDRDY = 0x0UL;
   uint32_t dummy = p_reg->EVENTS_TXDRDY;
   (void)dummy;
}

__STATIC_INLINE void fnClearErrorEvent(NRF_UART_Type * p_reg)
{
   p_reg->EVENTS_ERROR = 0x0UL;
   uint32_t dummy = p_reg->EVENTS_ERROR;
   (void)dummy;
}

__STATIC_INLINE void fnHandleReceive(void)
{
    uint32_t unNextHead = (gunReceiverHead + 1) & FIFO_WRAP;
    uint32_t unTail = gunReceiverTail;

    fnClearReceiveEvent(gpUartRegister);

    /*if there is no room in the fifo we will lose bytes ???*/
    if (unNextHead == unTail)
    {
        return; /*fifo full*/
    }

   uint8_t data = gpUartRegister->RXD;
   gaReceiveBuffer[gunReceiverHead] = data;
   gunReceiverHead = unNextHead;
}
/**
Check FTDI chip that it is ready to send.  
*/
__STATIC_INLINE void fnTxByte(void)
{
      uint8_t byTmp = gaTransmitBuffer[gunTransmitTail];
      gunTransmitTail++;
      gunTransmitTail &= FIFO_WRAP;
      gpUartRegister->TXD = byTmp; /*write to uart*/

}

__STATIC_INLINE void fnHandleTransmit(void)
{
   uint32_t unHead = gunTransmitHead;
   fnClearTransmitEvent(gpUartRegister);

   if (gunTransmitTail == unHead)
   {
      gbTxComplete = true;
   }
   else
   {
      fnTxByte();
   }
}

void UARTE0_UART0_IRQHandler(void)
{
   //TODO:: RXTO - Receiver Timeout
   if (fnIsErrorEvent(gpUartRegister))
   {
      LOG_DEBUG("UART error 0x%x\r\n", gpUartRegister->ERRORSRC);

      if (gpUartRegister->ERRORSRC & UART_ERRORSRC_BREAK_Msk)
      {
         LOG_DEBUG("UART break detected\r\n");
         gbIsBreakDetected = true;
      }
      else
      {
         const uint32_t bmErrors = (UART_ERRORSRC_FRAMING_Msk | UART_ERRORSRC_OVERRUN_Msk | UART_ERRORSRC_PARITY_Msk);
         gbIsErrorDetected = ((gpUartRegister->ERRORSRC & bmErrors) != 0) ? true : false;
      }

      fnClearErrorEvent(gpUartRegister);
   }
   else if (fnIsTransmitEvent(gpUartRegister))
   {
      fnHandleTransmit();
   }
   
   if (fnIsReceiveEvent(gpUartRegister))
   {
      fnHandleReceive();
   }
}